;; GNU Emacs code for Teletype 5620 mouse.
;; Copyright (C) Free Software Foundation Sept 1987.


;; This file is part of GNU Emacs.

;; Gnu Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it(se serves any particular purpose or works at all,
;; unless he says so in writing. Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.


;;; Author: paolo conti, September 1987 (conti@iis.uucp)
;;; ideas stolen from the version for the BBN Bitgraph mouse by John Robinson

(provide 'mega-lilith)

(defconst selection ?S
  "letter terminating selection events")

(defconst click ?R
  "letter terminating normal mouse events")

(defconst compressed-click ?Q
  "letter terminating compressed mouse events")

(defconst layer-resized ?C
  "letter terminating layer size events")

(defconst resizing-event ?c
  "letter terminating resizing events communicated autonomously by
   the terminal")

(defconst select 1
  "reported by terminal for a selection event")

(defconst copy 2
  "reported by terminal for a copy event")

(defconst delete 3
  "reported by terminal for a delete event")

(defconst move 4
  "reported by terminal for a move event")

(defun 5620-mouse-report ()
  "called every time a mouse event has happened; it decodes the event and executes the required actions"
  (interactive)
  (let* ((event (5620-get-mouse-event))
	 (type (car event))
	 (parameters (cdr event)))
    (cond
      ((eq type selection)
       (5620-handle-selection parameters))
      ((eq type click)
       (5620-handle-click parameters))
      ((eq type compressed-click)
       (5620-handle-compressed-click parameters))
      ((eq type layer-resized)
       (5620-resize-layer parameters))
      ((eq type resizing-event)
       (read-char)    ;; to flush the lf following the "c" in resizing events
       (5620-redraw-display)))))

(defun 5620-handle-selection (number-list)
  "checks wether the selected region is legal (i.e. in one window or
in two abutting windows on the same buffer); if so, executes the requested
action on the region (i.e. select, move, copy or delete)"
  (let* ((y0 (car number-list))
	 (x0 (nth 1 number-list))
	 (y1 (nth 2 number-list))	 
	 (x1 (nth 3 number-list))	 
	 (action (nth 4 number-list))
	 (window0 (5620-pos-to-window x0 y0))
	 (window1 (5620-pos-to-window x1 y1))
	 (edges0 (window-edges window0))
	 (edges1 (window-edges window1))
	 (top1-of-0 (eq (nth 3 edges0)
			(nth 1 edges1)))
	 (top0-of-1 (eq (nth 3 edges1)
		        (nth 1 edges0))))
    ;;; check for all possible invalid selections
    (cond ((not (or (eq window0 window1) top0-of-1 top1-of-0))
	   (error "invalid selection; windows don't abut"))
	  ((not (eq (window-buffer window0) (window-buffer window1)))
	   (error "invalid selection; windows not on same buffer"))
          ((or (eq y0 (screen-height)) (eq y1 (screen-height)))
           (error "invalid selection; selection ends in mini-buffer"))
	  ((or (eq y0 (nth 3 edges0))
	       (eq y1 (nth 3 edges1)))
	   (error "invalid selection; selection ends in mode line"))
	  ((or (= x0 0)
	       (= x1 0))
	   (error "invalid selection; selection ends in scroll bar"))
    ;;; treat a valid selection
	  (t
	   (let ((old-window (selected-window))
		 (old-point (point-marker))
		 (window-x0 (1- (- x0 (nth 0 edges0))))
		 (window-y0 (1- (- y0 (nth 1 edges0))))
		 (window-x1 (1- (- x1 (nth 0 edges1))))
		 (window-y1 (1- (- y1 (nth 1 edges1))))
		 end-of-selection
		 point-in-new-window)
	     ;find and save end of selection
	     (select-window window1)
	     (5620-move-point-to-x-y window-x1 window-y1)
	     (setq end-of-selection (point))
	     ;go to start of selection
	     (select-window window0)
	     (setq point-in-new-window (point))
	     (5620-move-point-to-x-y window-x0 window-y0)
	     ;treat according to action
	     (cond
	      ((/= end-of-selection (point))
	       (cond
		((or (eq action select) (eq action copy))
		 ; avoid appending in previous kill entry
		 (setq last-command nil)
		 (copy-region-as-kill (point) end-of-selection))
		; case action move or delete
		(t
		 (setq last-command nil)
		 (kill-region (point) end-of-selection)))
	       (cond
		((or (eq action copy) (eq action move))
		 (select-window old-window)
		 (goto-char (marker-position old-point))
		 (setq this-command 'yank)
		 (yank)
		 (pop-mark)
		 (princ ""))
		((eq action select)
		 (goto-char point-in-new-window)
		 (select-window old-window)
		 (goto-char (marker-position old-point)))))
	      ;;; selection is empty;
	      (t
	       (push-mark end-of-selection)
	       (goto-char point-in-new-window)
	       (select-window old-window)
	       (goto-char (marker-position old-point)))))))))


;;; the following information may be accessed by the functions
;;; bound to mouse events; these functions all don't take parameters
;;; but access global variables

(defvar buttons0 0
  "buttons pressed for the current click event")

(defvar x0
  "screen column where a mouse button was last pressed")

(defvar y0
  "screen line where a mouse button was last pressed")

(defvar window0
  "window in which a mouse button was last pressed")

(defvar site0
  "site in which a mouse button was last pressed; sites include the scroll-bar,
the mode-line etc")

(defvar x1
  "screen column where the mouse button(s) was (were) released")

(defvar y1
  "screen line where the mouse button(s) was (were) released")


(defun 5620-handle-click (parameters)
  "handles the mouse events which are not selection events; calls the function
bound to the given region of the screen and to the givenn buttons"
  (let ((new-buttons (nth 2 parameters)))
    (cond
     ((= buttons0 0)
      (setq buttons0 new-buttons)
      (setq y0 (car parameters))
      (setq x0 (nth 1 parameters))
      (let ((site-info (5620-get-site-info x0 y0)))
	(setq site0 (car site-info))
	(setq window0 (nth 1 site-info))))
     ((> new-buttons buttons0)
      (setq buttons0 new-buttons))
     ((= new-buttons 0)
      (setq y1 (car parameters))
      (setq x1 (nth 1 parameters))
      (let ((action (aref (aref 5620-button-actions site0)
			  (1- buttons0))))
	(setq buttons0 0)
	(cond
	 ((not (null action))
	  (funcall action))
	 (t
	  (error "Gaat's no? No action defined for those buttons")))))
     ;;; some, but not all buttons released
     (t nil))))


(defun 5620-handle-compressed-click (parameters)
  "handles the mouse events which are not selection events; calls the function
bound to the given region of the screen and to the given buttons"
    (setq y0 (car parameters))
    (setq x0 (nth 1 parameters))
    (setq y1 (nth 2 parameters))
    (setq x1 (nth 3 parameters))
    (setq buttons0 (nth 4 parameters))
    (let ((site-info (5620-get-site-info x0 y0)))
      (setq site0 (car site-info))
      (setq window0 (nth 1 site-info)))
    (let ((action (aref (aref 5620-button-actions site0)
			  (1- buttons0))))
      (setq buttons0 0)
      (cond
	((not (null action))
	 (funcall action))
	(t
	  (error "Gaat's no? No action defined for those buttons")))))

(defun 5620-get-site-info (x y)
  "returns a list with the site (scroll-bar, buffer-area etc) and with the
window the coordinates lie in"
  (let* ((window (5620-pos-to-window x y))
	 (edges (window-edges window))
	 (site
	  (cond
	   ((= y (screen-height))
	    (cond
	     ((= x 0)
	      echo-close-box)
	     (t
              echo-area)))
	   ((= x 0)
	    (cond
	     ((= y (nth 3 edges))
	      close-box)
	     (t scroll-bar)))
	   ((= y (nth 3 edges))
	    mode-line)
	   (t buffer-area))))
    (list site window)))

;;;
;;; functions which may be bound to clicks in different sites
;;;

(defun 5620-scroll-up ()
  "scrolls the line the mouse was clicked in to the top of its window" 
  (let ((site-info (5620-get-site-info x1 y1))
	(current-window (selected-window)))
    (cond
     ((and (eq (car site-info) scroll-bar)
	   (eq (nth 1 site-info) window0))
      (select-window window0)
      (scroll-up (- y1 1 (nth 1 (window-edges window0))))
      (select-window current-window)))))

(defconst 5620-symmetric-scrolling nil
  "*if nil, the action 5620-scroll-down is the perfect inverse of
5620-scroll-up; otherwise the pointed line is scrolled to the bottom")

(defun 5620-scroll-down ()
  "scrolls down in the window the mouse was clicked in; the semantic depends
 on the value of the variable 5620-scroll-line-to-bottom"
  (let ((site-info (5620-get-site-info x1 y1))
	(current-window (selected-window)))
    (cond
     ((and (eq (car site-info) scroll-bar)
	   (eq (nth 1 site-info) window0))
      (select-window window0)
      (cond
       ((null 5620-symmetric-scrolling)
	(scroll-up (- (1+ y1)  (nth 3 (window-edges window0)))))
       (t (scroll-up (- (nth 1 (window-edges window0)) (1- y1)))))
      (select-window current-window)))))

(defun 5620-set-point ()
  "selects the window where the mouse was clicked and sets it's point at
the proper location"
  (let ((site-info (5620-get-site-info x1 y1)))
    (cond ((and (eq (car site-info) buffer-area)
	        (eq (nth 1 site-info) window0))
	   (select-window window0)
	   (let ((edges0 (window-edges window0)))
	     (5620-move-point-to-x-y
	      (1- (- x1 (nth 0 edges0)))
	      (1- (- y1 (nth 1 edges0)))))))))

(defun 5620-yank ()
  "inserts the top of the kill ring at point (and not at the location of
the mouse click!); if the mouse was not clicked in the selected-window,
the function does nothing"
  (let ((site-info (5620-get-site-info x1 y1)))
    (cond
     ((and (eq (car site-info) buffer-area)
	   (eq (nth 1 site-info) window0)
	   (eq window0 (selected-window)))
      (setq this-command 'yank)
      (yank)))))

(defun 5620-split-window ()
  "splits the window in which the mouse was clicked at the line the click
occured"
  (let ((site-info (5620-get-site-info x1 y1))
	(current-window (selected-window)))
    (cond
     ((and (eq (car site-info) scroll-bar)
	   (eq (nth 1 site-info) window0))
      (select-window window0)
      (let ((edges0 (window-edges window0)))
	(split-window-vertically (- y1 (nth 1 edges0)))
	(select-window current-window))))))

(defun 5620-move-border ()
  "moves the border on which the mouse was pressed to the line the mouse
was released. Allows to enlarge a window at the cost of its neighbour.
The neighbour window can't be killed with 5620-move-border. The echo area
can't be enlarged with 5620-move-border."
    (cond
     ((and (eq site0 mode-line)
	   (/= y0 (1- (screen-height))))
      (let* ((current-window (selected-window))
	     (edges0 (window-edges window0))
	     (offset (- y1 (nth 3 edges0)))
	     (enable
	      (cond
	       ((< offset 0)
		(>= offset (- window-min-height (window-height window0))))
	       ((> offset 0)
		(<= offset
		    (- (window-height (5620-pos-to-window x0 (1+ y0)))
		       window-min-height))))))
	(cond
	 (enable
	  (select-window window0)
	  (enlarge-window offset)
	  (cond
	   ((window-point current-window)
	    (select-window current-window)))))))))

(defun 5620-flip-from-column ()
  "displays in the window the part of the buffer corresponding to the ratio
between the column where the mouse was clicked and the number of columns of
the screen. Flipping in the first column displays the top of the buffer, flipping
in the central column the central part of the buffer etc"
  (let ((site-info (5620-get-site-info x1 y1))
	(current-window (selected-window)))
    (cond
     ((and (eq (car site-info) site0)
	   (eq (nth 1 site-info) window0))
      (select-window window0)

      ;; we'd like to have
      ;; (goto-char (+ (point-min)
      ;; 	     (/ (* (1- x1)
      ;; 	       (- (point-max) (point-min)))
      ;;	    (1- (window-width)))))
      ;; but we must avoid the first multiplication in order to avoid
      ;; an overflow for very long files;
       (let* ((a (1- x1))
	      (b (- (point-max) (point-min)))
	      (c (1- (window-width)))
	      (a_div_c (/ a c))
	      (a_mod_c (% a c))
	      (b_div_c (/ b c))
	      (b_mod_c (% b c)))
	 (goto-char ( + (point-min)
			(+ (* c a_div_c b_div_c)
			   (* a_div_c b_mod_c)
			   (* a_mod_c b_div_c)
			   (/ (* a_mod_c b_mod_c) c)))))
      (select-window current-window)))))

(defun 5620-flip-from-line ()
  "displays in the window the part of the buffer corresponding to the ratio
between the window line where the mouse was clicked and the number of lines of
the screen. Flipping in the top line displays the top of the buffer, flipping
in the middle of the window the central part of the buffer etc
Note: 5620-flip-from-column will usually permit a more precise flipping, since most windows have much more columns than lines"
  (let ((site-info (5620-get-site-info x1 y1))
	(current-window (selected-window)))
    (cond
     ((and (eq (car site-info) site0)
	   (eq (nth 1 site-info) window0))
      (select-window window0)

      ;; we'd like to have
      ;; (goto-char ( + (point-min)
      ;;	     (/ (* (1- (- y1 (nth 1 (window-edges))))
      ;;	          (- (point-max) (point-min)))
      ;;	       (-  (window-height) 2))))
      ;; but we must avoid the first multiplication in order to avoid
      ;; an overflow for very long files;
       (let* ((a (1- (- y1 (nth 1 (window-edges)))))
	      (b (- (point-max) (point-min)))
	      (c (-  (window-height) 2))
	      (a_div_c (/ a c))
	      (a_mod_c (% a c))
	      (b_div_c (/ b c))
	      (b_mod_c (% b c)))
	 (goto-char ( + (point-min)
			(+ (* c a_div_c b_div_c)
			   (* a_div_c b_mod_c)
			   (* a_mod_c b_div_c)
			   (/ (* a_mod_c b_mod_c) c)))))
       (select-window current-window)))))

(defun 5620-select-window ()
  "makes the window the mouse was clicked in the active window"
  (let ((site-info (5620-get-site-info x1 y1))
	(current-window (selected-window)))
    (cond
     ((and (eq (car site-info) site0)
	   (eq (nth 1 site-info) window0))
      (select-window window0)))))

(defun 5620-delete-window ()
  "kills the window the mouse was clicked in"
  (let ((site-info (5620-get-site-info x1 y1))
	(current-window (selected-window)))
    (cond
     ((and (eq (car site-info) site0)
	   (eq (nth 1 site-info) window0))
      (delete-window window0)
      (cond
       ((window-point current-window)
	(select-window current-window)))))))

(defun 5620-select-minibuffer ()
  "Selects the minibuffer if it is active"
  (let ((edges (window-edges))
	(window nil))
    (while (and (not (eq window (selected-window)))
		(>  (screen-height) (nth 3 edges)))
      (setq window (next-window window))
      (setq edges (window-edges window)))
    (select-window window)))

(defun 5620-find-forward ()
  "Searches forward for the string in the kill ring"
  (let ((search-string (car kill-ring)))
    (cond
     ((looking-at search-string)
      (search-forward search-string)))
    (search-forward search-string)))

(defun 5620-find-backward ()
  "Searches backward for the string in the kill ring."
  (let ((search-string (car kill-ring))
	(old-point (point)))
    (search-backward search-string)
    (cond
     ((eq (length search-string)
	  (- old-point (point)))
      (search-backward search-string)))))

(defun 5620-set-sam-scrolling ()
  "Binds the buttons in the scroll-bar the same way the editor sam does"
  (interactive)
  (5620-bind-click 'right 'scroll-bar '5620-scroll-up)
  (5620-bind-click 'left 'scroll-bar '5620-scroll-down)
  (5620-bind-click 'middle 'scroll-bar '5620-flip-from-line)
  (5620-bind-click 'left-middle 'scroll-bar '5620-split-window))

(defun 5620-set-lilith-scrolling ()
  "Binds the buttons in the scroll-bar the default way"
  (interactive)
  (5620-bind-click 'left 'scroll-bar '5620-scroll-up)
  (5620-bind-click 'right 'scroll-bar '5620-scroll-down)
  (5620-bind-click 'middle 'scroll-bar '5620-split-window))


(defconst buffer-area 0
  "*the whole Emacs window except its mode line")

(defconst scroll-bar 1
  "*the leftmost column of the screen")

(defconst mode-line 2
  "*the lowest line of an Emacs window (inverted)")

(defconst close-box 3
  "*the part of the scroll bar at the left of the mode line")

(defconst echo-area 4
  "*the bottom line of the screen or layer")

(defconst echo-close-box 5
  "*the part of the scroll bar at the left of the echo area") 

(defun make-vector-7 ()
  (vector nil nil nil nil nil nil nil))

(defvar 5620-button-actions
  (vector (make-vector-7) (make-vector-7) (make-vector-7)
	  (make-vector-7) (make-vector-7) (make-vector-7)))


(defconst right             1 "*")
(defconst middle            2 "*")
(defconst middle-right      3 "*")
(defconst left              4 "*")
(defconst left-right        5 "*")
(defconst left-middle       6 "*")
(defconst left-middle-right 7 "*")

(defun 5620-bind-click (buttons-symbol site-symbol action)
  "Binds BUTTONS clicked in SITE to the function ACTION"
  (interactive "vButtons: \nvSite: \naAction: ")
  (let((site (eval site-symbol))
       (buttons (eval buttons-symbol)))
    (if (and (integerp buttons) (> buttons 0) (< buttons 8))
	(cond
	 ((< site (length 5620-button-actions))
	  (aset (aref 5620-button-actions site) (1- buttons) action))
	 (t
	  (error (concat site-symbol " is not a predefined site"))))
      (error "buttons should be between 1 and 7"))))

(5620-bind-click 'left 'scroll-bar '5620-scroll-up)
(5620-bind-click 'right 'scroll-bar '5620-scroll-down)
(5620-bind-click 'middle 'scroll-bar '5620-split-window)
(5620-bind-click 'left 'buffer-area '5620-set-point)
(5620-bind-click 'middle 'buffer-area '5620-yank)
(5620-bind-click 'left 'close-box '5620-select-window)
(5620-bind-click 'left-right 'close-box '5620-delete-window)
(5620-bind-click 'middle 'mode-line '5620-move-border)
(5620-bind-click 'left 'mode-line '5620-flip-from-column)
(5620-bind-click 'left 'echo-area '5620-find-forward)
(5620-bind-click 'middle 'echo-area '5620-find-backward)
(5620-bind-click 'left 'echo-close-box '5620-select-minibuffer)

(defun 5620-bind-selection-buttons (select-button copy-button)
  "sets the SELECT-BUTTON and COPY-BUTTON; these must be different and
both must be in {left, middle, right}; the delete button is set automatically
to the third button"
  (interactive "vSelection Button: \nvCopy Button: ")
  (setq select-button (eval select-button))
  (setq copy-button (eval copy-button))
  (cond
   ((/= select-button copy-button)
    (send-string-to-terminal
     (concat "\e[?" 0 ";" select-button ";" copy-button "b")))
   (t
    (error "selection and copy button must be different"))))


(defun 5620-get-tty-num ()
  "Read an integer from terminal, and return a list with the char following
   the int and the int"
  (let
      ((num 0)
       (char (- (read-char) 48)))
    (while (and (>= char 0)
		(<= char 9))
      (setq num (+ (* num 10) char))
      (setq char (- (read-char) 48)))
    (list (+ char 48) num)))

(defun 5620-get-mouse-event ()
  "Read a list of numbers from the terminal until one of them is not
   terminated by a semi-colon and reports a list containing the numbers read"
  (interactive)
  (let ((result nil)
	(next-item (5620-get-tty-num)))
    (while
	(= (car next-item) ?;)
      (setq result (append result (cdr next-item)))
      (setq next-item (5620-get-tty-num)))
    (setq result (append result (cdr next-item)))
    (cons (car next-item) result)))

(defun 5620-move-point-to-x-y (x y)
  "Position cursor in window coordinates.
X and Y are 0-based character positions in the window."
  (move-to-window-line y)
  (move-to-column (+ (current-column)
		     (min (- (window-width) 1) x))))

(defun 5620-pos-to-window (x y)
  "Find window corresponding to screen coordinates.
X and Y are 1-based character positions on the screen."
  (let ((edges (window-edges))
	(window nil))
    (setq x (max x 1))
    (while (and (not (eq window (selected-window)))
		(or (<= y (nth 1 edges))
		    (>  y (nth 3 edges))
		    (<= x (nth 0 edges))
		    (>  x (nth 2 edges))))
      (setq window (next-window window))
      (setq edges (window-edges window)))
    (or window (selected-window))))

(defvar redraw-called-interacively nil
  "tells wether 5620-redraw-display was called from the user, in which case
the screen is redrawn at least once, or wether it was called after the
screen size has been required, in which case the screen is redrawn only if
it's size has changed; the solution with the global variable is very ugly")

(defun 5620-redraw-display ()
  "refreshes the screen after having asked how big the layer currently is"
  (interactive)
  (setq redraw-called-interacively t)
  (send-string-to-terminal "\e[?C"))

(defun 5620-resize-layer (parameters)
  (let ((new-height (nth 0 parameters))
	(new-width (1- (nth 1 parameters)))
	layer-resized)
    (cond
     ((/= new-height (screen-height))
      (set-screen-height new-height)
      (setq layer-resized t)))
    (cond
     ((/= new-width (screen-width))
      (set-screen-width new-width)
      (setq layer-resized t)))
    (cond
     ((and
       redraw-called-interacively
       (null layer-resized))
      (redraw-display)))
    (setq redraw-called-interacively nil)))

(setq suspend-hook
      (cond
       ((and (boundp 'suspend-hook) suspend-hook)
	(append suspend-hook '((send-string-to-terminal "\e[?0s"))))
       (t
        '(lambda ()
	  (send-string-to-terminal "\e[?0s")))))

(setq suspend-resume-hook
      (cond
       ((and (boundp 'suspend-resume-hook) suspend-resume-hook)
	(append suspend-resume-hook '((send-string-to-terminal "\e[?1s")
				      (send-string-to-terminal "\e[?C"))))
       (t
        '(lambda ()
	   (send-string-to-terminal "\e[?1s")
	   (send-string-to-terminal "\e[?C")))))

;;;
;;; send initialization strings to the terminal
;;;

;;; enable mouse
(send-string-to-terminal "\e[?1m")

;;; enable scroll-bar
(send-string-to-terminal "\e[?1s")

;;; take scroll bar in account
(set-screen-width (1- (screen-width)))

;;;define default selection and copy buttons
(5620-bind-selection-buttons 'right 'middle)

;;; unbind "\e[" if it is bound to a function, since otherwise
;;; "\e[" can't be bound

(let ((binding (global-key-binding "\e[")))
  (cond
   ((and (not (null binding))
	 (not (keymapp binding)))
    (global-unset-key "\e["))))
;;
;;
;;
(global-set-key "\e[?" '5620-mouse-report)
;;
;;; ask for current layer size
;;
(send-string-to-terminal "\e[?C")
;;
;;
;;
(run-hooks 'mega-lilith-hook)
